from time import time
from random import random
from heapq import heappush, heappop

from .lmovegen import genAllMoves, genCheckEvasions, genCaptures
from .egtb_gaviota import EgtbGaviota
from pychess.Utils.const import (
    ATOMICCHESS,
    KINGOFTHEHILLCHESS,
    THREECHECKCHESS,
    LOSERSCHESS,
    SUICIDECHESS,
    GIVEAWAYCHESS,
    EMPTY,
    PROMOTIONS,
    DROP,
    KING,
    RACINGKINGSCHESS,
    hashfALPHA,
    hashfBETA,
    hashfEXACT,
    hashfBAD,
    DRAW,
    WHITE,
    WHITEWON,
)
from .leval import evaluateComplete
from .lsort import getCaptureValue, getMoveValue
from .ldata import MATE_VALUE, VALUE_AT_PLY
from .TranspositionTable import TranspositionTable
from pychess.Variants.atomic import kingExplode
from pychess.Variants.kingofthehill import testKingInCenter
from pychess.Variants.suicide import pieceCount
from pychess.Variants.threecheck import checkCount
from . import ldraw

TIMECHECK_FREQ = 500

table = TranspositionTable(32 * 1024 * 1024)
skipPruneChance = 0
searching = False
nodes = 0
endtime = 0
timecheck_counter = TIMECHECK_FREQ
egtb = None


def alphaBeta(board, depth, alpha=-MATE_VALUE, beta=MATE_VALUE, ply=0):
    """This is a alphabeta/negamax/quiescent/iterativedeepend search algorithm
    Based on moves found by the validator.py findmoves2 function and
    evaluated by eval.py.
    The function recalls itself "depth" times. If the last move in range
    depth was a capture, it will continue calling itself, only searching for
    captures.
    It returns a tuple of
    *   a list of the path it found through the search tree (last item being
        the deepest)
    *   a score of your standing the the last possition."""

    global searching, nodes, table, endtime, timecheck_counter
    foundPv = False
    hashf = hashfALPHA
    amove = []

    ############################################################################
    # Mate distance pruning
    ############################################################################

    MATED = -MATE_VALUE + ply
    MATE_IN_1 = MATE_VALUE - ply - 1

    if beta <= MATED:
        return [], MATED
    if beta >= MATE_IN_1:
        beta = MATE_IN_1
        if alpha >= beta:
            return [], MATE_IN_1

    if board.variant == ATOMICCHESS:
        if bin(board.boards[board.color][KING]).count("1") == 0:
            return [], MATED
    elif board.variant == LOSERSCHESS:
        if pieceCount(board, board.color) == 1:
            return [], -MATED
    elif board.variant == SUICIDECHESS or board.variant == GIVEAWAYCHESS:
        if pieceCount(board, board.color) == 0:
            return [], -MATED
    elif board.variant == KINGOFTHEHILLCHESS:
        if testKingInCenter(board):
            return [], MATED
    elif board.variant == THREECHECKCHESS:
        if checkCount(board, board.color) == 3:
            return [], MATED

    ############################################################################
    # Look in the end game table
    ############################################################################

    global egtb
    if egtb:
        tbhits = egtb.scoreAllMoves(board)
        if tbhits:
            move, state, steps = tbhits[0]

            if state == DRAW:
                score = 0
            elif board.color == WHITE:
                if state == WHITEWON:
                    score = MATE_VALUE - steps
                else:
                    score = -MATE_VALUE + steps
            else:
                if state == WHITEWON:
                    score = -MATE_VALUE + steps
                else:
                    score = MATE_VALUE - steps
            return [move], score

    ###########################################################################
    # We don't save repetition in the table, so we need to test draw before   #
    # table.                                                                  #
    ###########################################################################

    # We don't adjudicate draws. Clients may have different rules for that.
    if ply > 0:
        if ldraw.test(board):
            return [], 0

    ############################################################################
    # Look up transposition table                                              #
    ############################################################################
    if ply == 0:
        table.newSearch()

    table.setHashMove(depth, -1)
    probe = table.probe(board, depth, alpha, beta)
    if probe:
        move, score, hashf = probe
        score = VALUE_AT_PLY(score, ply)
        table.setHashMove(depth, move)

        if hashf == hashfEXACT:
            return [move], score
        elif hashf == hashfBETA:
            beta = min(score, beta)
        elif hashf == hashfALPHA:
            alpha = score

        if hashf != hashfBAD and alpha >= beta:
            return [move], score

    ############################################################################
    # Cheking the time                                                         #
    ############################################################################

    timecheck_counter -= 1
    if timecheck_counter == 0:
        if time() > endtime:
            searching = False
        timecheck_counter = TIMECHECK_FREQ

    ############################################################################
    # Break itereation if interupted or if times up                            #
    ############################################################################

    if not searching:
        return [], -evaluateComplete(board, 1 - board.color)

    ############################################################################
    # Go for quiescent search                                                  #
    ############################################################################

    isCheck = board.isChecked()

    if depth <= 0:
        if isCheck:
            # Being in check is that serious, that we want to take a deeper look
            depth += 1
        elif board.variant in (
            LOSERSCHESS,
            SUICIDECHESS,
            GIVEAWAYCHESS,
            ATOMICCHESS,
            RACINGKINGSCHESS,
        ):
            return [], evaluateComplete(board, board.color)
        else:
            mvs, val = quiescent(board, alpha, beta, ply)
            return mvs, val

    ############################################################################
    # Find and sort moves                                                      #
    ############################################################################

    if board.variant in (LOSERSCHESS, SUICIDECHESS, GIVEAWAYCHESS):
        mlist = [m for m in genCaptures(board)]
        if board.variant == LOSERSCHESS:
            if isCheck:
                evasions = [m for m in genCheckEvasions(board)]
                eva_cap = [m for m in evasions if m in mlist]
                mlist = eva_cap if eva_cap else evasions
            else:
                valid_captures = []
                for move in mlist:
                    board.applyMove(move)
                    if not board.opIsChecked():
                        valid_captures.append(move)
                    board.popMove()
                mlist = valid_captures
        if not mlist and not isCheck:
            mlist = [m for m in genAllMoves(board)]
        moves = [(-getMoveValue(board, table, depth, m), m) for m in mlist]
    elif board.variant == ATOMICCHESS:
        if isCheck:
            mlist = [
                m
                for m in genCheckEvasions(board)
                if not kingExplode(board, m, board.color)
            ]
        else:
            mlist = [
                m for m in genAllMoves(board) if not kingExplode(board, m, board.color)
            ]
        moves = [(-getMoveValue(board, table, depth, m), m) for m in mlist]
    elif board.variant == RACINGKINGSCHESS:
        mlist = [m for m in genAllMoves(board) if not board.willGiveCheck(m)]
        moves = [(-getMoveValue(board, table, depth, m), m) for m in mlist]
    else:
        if isCheck:
            moves = [
                (-getMoveValue(board, table, depth, m), m)
                for m in genCheckEvasions(board)
            ]
        else:
            moves = [
                (-getMoveValue(board, table, depth, m), m) for m in genAllMoves(board)
            ]
    moves.sort()

    # This is needed on checkmate
    catchFailLow = None

    ############################################################################
    # Loop moves                                                               #
    ############################################################################

    for moveValue, move in moves:
        nodes += 1

        board.applyMove(move)
        if not isCheck:
            if board.opIsChecked():
                board.popMove()
                continue

        catchFailLow = move

        if foundPv:
            mvs, val = alphaBeta(board, depth - 1, -alpha - 1, -alpha, ply + 1)
            val = -val
            if val > alpha and val < beta:
                mvs, val = alphaBeta(board, depth - 1, -beta, -alpha, ply + 1)
                val = -val
        else:
            mvs, val = alphaBeta(board, depth - 1, -beta, -alpha, ply + 1)
            val = -val

        board.popMove()

        if val > alpha:
            if val >= beta:
                if searching and move >> 12 != DROP:
                    table.record(
                        board, move, VALUE_AT_PLY(beta, -ply), hashfBETA, depth
                    )
                    # We don't want to use our valuable killer move spaces for
                    # captures and promotions, as these are searched early anyways.
                    if (
                        board.arBoard[move & 63] == EMPTY
                        and not move >> 12 in PROMOTIONS
                    ):
                        table.addKiller(depth, move)
                        table.addButterfly(move, depth)
                return [move] + mvs, beta

            alpha = val
            amove = [move] + mvs
            hashf = hashfEXACT
            foundPv = True

    ############################################################################
    # Return                                                                   #
    ############################################################################

    if amove:
        if searching:
            table.record(board, amove[0], VALUE_AT_PLY(alpha, -ply), hashf, depth)
            if board.arBoard[amove[0] & 63] == EMPTY:
                table.addKiller(depth, amove[0])
        return amove, alpha

    if catchFailLow:
        if searching:
            table.record(board, catchFailLow, VALUE_AT_PLY(alpha, -ply), hashf, depth)
        return [catchFailLow], alpha

    # If no moves were found, this must be a mate or stalemate
    if isCheck:
        return [], MATED

    return [], 0


def quiescent(board, alpha, beta, ply):
    if skipPruneChance and random() < skipPruneChance:
        return [], (alpha + beta) // 2

    global searching, nodes, endtime, timecheck_counter

    if ldraw.test(board):
        return [], 0

    timecheck_counter -= 1
    if timecheck_counter == 0:
        if time() > endtime:
            searching = False
        timecheck_counter = TIMECHECK_FREQ

    ############################################################################
    # Break itereation if interupted or if times up                            #
    ############################################################################

    if not searching:
        return [], -evaluateComplete(board, 1 - board.color)

    isCheck = board.isChecked()

    # no stand-pat when in check
    if not isCheck:
        value = evaluateComplete(board, board.color)
        if value >= beta:
            return [], beta
        if value > alpha:
            alpha = value

    amove = []

    heap = []

    if isCheck:
        someMove = False
        for move in genCheckEvasions(board):
            someMove = True
            # Heap.append is fine, as we don't really do sorting on the few moves
            heap.append((0, move))
        if not someMove:
            return [], -MATE_VALUE + ply
    else:
        for move in genCaptures(board):
            heappush(heap, (-getCaptureValue(board, move), move))

    while heap:
        nodes += 1

        v, move = heappop(heap)

        board.applyMove(move)
        if not isCheck:
            if board.opIsChecked():
                board.popMove()
                continue

        mvs, val = quiescent(board, -beta, -alpha, ply + 1)
        val = -val

        board.popMove()

        if val >= beta:
            return [move] + mvs, beta

        if val > alpha:
            alpha = val
            amove = [move] + mvs

    if amove:
        return amove, alpha

    else:
        return [], alpha


class EndgameTable:
    def __init__(self):
        self.provider = EgtbGaviota()

    def _pieceCounts(self, board):
        return sorted([bin(board.friends[i]).count("1") for i in range(2)])

    def scoreAllMoves(self, lBoard):
        """Return each move's result and depth to mate.
        lBoard: A low-level board structure
        Return value: a list, with best moves first, of:
        move: A high-level move structure
        game_result: Either WHITEWON, DRAW, BLACKWON
        depth: Depth to mate
        """

        pc = self._pieceCounts(lBoard)
        if self.provider.supports(pc):
            return self.provider.scoreAllMoves(lBoard)
        return []


def enableEGTB():
    global egtb
    egtb = EndgameTable()



"""Monkey patching asyncio.StreamReader to add readuntil() from Python 3.5.2"""

import asyncio


class IncompleteReadError(EOFError):
    """
    Incomplete read error. Attributes:

    - partial: read bytes string before the end of stream was reached
    - expected: total number of expected bytes (or None if unknown)
    """

    def __init__(self, partial, expected):
        super().__init__(
            "%d bytes read on a total of %r expected bytes" % (len(partial), expected)
        )
        self.partial = partial
        self.expected = expected


class LimitOverrunError(Exception):
    """Reached the buffer limit while looking for a separator.

    Attributes:
    - consumed: total number of to be consumed bytes.
    """

    def __init__(self, message, consumed):
        super().__init__(message)
        self.consumed = consumed


async def _wait_for_data(self, func_name):
    """Wait until feed_data() or feed_eof() is called.

    If stream was paused, automatically resume it.
    """
    # StreamReader uses a future to link the protocol feed_data() method
    # to a read coroutine. Running two read coroutines at the same time
    # would have an unexpected behaviour. It would not possible to know
    # which coroutine would get the next data.
    if self._waiter is not None:
        raise RuntimeError(
            "%s() called while another coroutine is "
            "already waiting for incoming data" % func_name
        )

    assert not self._eof, "_wait_for_data after EOF"

    # Waiting for data while paused will make deadlock, so prevent it.
    if self._paused:
        self._paused = False
        self._transport.resume_reading()

    self._waiter = asyncio.futures.Future(loop=self._loop)
    try:
        await self._waiter
    finally:
        self._waiter = None


async def readuntil(self, separator=b"\n"):
    """Read data from the stream until ``separator`` is found.

    On success, the data and separator will be removed from the
    internal buffer (consumed). Returned data will include the
    separator at the end.

    Configured stream limit is used to check result. Limit sets the
    maximal length of data that can be returned, not counting the
    separator.

    If an EOF occurs and the complete separator is still not found,
    an IncompleteReadError exception will be raised, and the internal
    buffer will be reset.  The IncompleteReadError.partial attribute
    may contain the separator partially.

    If the data cannot be read because of over limit, a
    LimitOverrunError exception  will be raised, and the data
    will be left in the internal buffer, so it can be read again.
    """
    seplen = len(separator)
    if seplen == 0:
        raise ValueError("Separator should be at least one-byte string")

    if self._exception is not None:
        raise self._exception

    # Consume whole buffer except last bytes, which length is
    # one less than seplen. Let's check corner cases with
    # separator='SEPARATOR':
    # * we have received almost complete separator (without last
    #   byte). i.e buffer='some textSEPARATO'. In this case we
    #   can safely consume len(separator) - 1 bytes.
    # * last byte of buffer is first byte of separator, i.e.
    #   buffer='abcdefghijklmnopqrS'. We may safely consume
    #   everything except that last byte, but this require to
    #   analyze bytes of buffer that match partial separator.
    #   This is slow and/or require FSM. For this case our
    #   implementation is not optimal, since require rescanning
    #   of data that is known to not belong to separator. In
    #   real world, separator will not be so long to notice
    #   performance problems. Even when reading MIME-encoded
    #   messages :)

    # `offset` is the number of bytes from the beginning of the buffer
    # where there is no occurrence of `separator`.
    offset = 0

    # Loop until we find `separator` in the buffer, exceed the buffer size,
    # or an EOF has happened.
    while True:
        buflen = len(self._buffer)

        # Check if we now have enough data in the buffer for `separator` to
        # fit.
        if buflen - offset >= seplen:
            isep = self._buffer.find(separator, offset)

            if isep != -1:
                # `separator` is in the buffer. `isep` will be used later
                # to retrieve the data.
                break

            # see upper comment for explanation.
            offset = buflen + 1 - seplen
            if offset > self._limit:
                raise LimitOverrunError(
                    "Separator is not found, and chunk exceed the limit", offset
                )

        # Complete message (with full separator) may be present in buffer
        # even when EOF flag is set. This may happen when the last chunk
        # adds data which makes separator be found. That's why we check for
        # EOF *ater* inspecting the buffer.
        if self._eof:
            chunk = bytes(self._buffer)
            self._buffer.clear()
            raise IncompleteReadError(chunk, None)

        # _wait_for_data() will resume reading if stream was paused.
        await self._wait_for_data("readuntil")

    if isep > self._limit:
        raise LimitOverrunError(
            "Separator is found, but chunk is longer than limit", isep
        )

    chunk = self._buffer[: isep + seplen]
    del self._buffer[: isep + seplen]
    self._maybe_resume_transport()
    return bytes(chunk)



from gi.repository import Gtk, GdkPixbuf

from pychess.ic.FICSObjects import FICSChallenge

SEPARATOR, ACCEPT, ASSESS, OBSERVE, FOLLOW, CHAT, CHALLENGE, FINGER, ARCHIVED = range(9)


def cmp(x, y):
    return (x > y) - (x < y)


class ParrentListSection:
    """Parrent for sections mainly consisting of a large treeview"""

    def button_press_event(self, treeview, event):
        if event.button == 3:  # right click
            pathinfo = treeview.get_path_at_pos(int(event.x), int(event.y))
            if pathinfo is not None:
                path, col = pathinfo[0], pathinfo[1]
                treeview.grab_focus()
                treeview.set_cursor(path, col, 0)
                self.menu.show_all()
                self.menu.popup(
                    None, None, None, None, event.button, Gtk.get_current_event_time()
                )
            return True
        return False

    def createLocalMenu(self, items):
        ITEM_MAP = {
            ACCEPT: (_("Accept"), self.on_accept),
            ASSESS: (_("Assess"), self.on_assess),
            OBSERVE: (_("Observe"), self.on_observe),
            FOLLOW: (_("Follow"), self.on_follow),
            CHAT: (_("Chat"), self.on_chat),
            CHALLENGE: (_("Challenge"), self.on_challenge),
            FINGER: (_("Finger"), self.on_finger),
            ARCHIVED: (_("Archived"), self.on_archived),
        }

        self.menu = Gtk.Menu()
        for item in items:
            if item == SEPARATOR:
                menu_item = Gtk.SeparatorMenuItem()
            else:
                label, callback = ITEM_MAP[item]
                menu_item = Gtk.MenuItem(label)
                menu_item.connect("activate", callback)
            self.menu.append(menu_item)
        self.menu.attach_to_widget(self.tv, None)

    def addColumns(self, treeview, *columns, **keyargs):
        if "hide" in keyargs:
            hide = keyargs["hide"]
        else:
            hide = []
        if "pix" in keyargs:
            pix = keyargs["pix"]
        else:
            pix = []
        for i, name in enumerate(columns):
            if i in hide:
                continue
            if i in pix:
                crp = Gtk.CellRendererPixbuf()
                crp.props.xalign = 0.5
                column = Gtk.TreeViewColumn(name, crp, pixbuf=i)
            else:
                crt = Gtk.CellRendererText()
                column = Gtk.TreeViewColumn(name, crt, text=i)
                column.set_resizable(True)
            column.set_sort_column_id(i)
            # prevent columns appear choppy
            column.set_sizing(Gtk.TreeViewColumnSizing.GROW_ONLY)

            column.set_reorderable(True)
            treeview.append_column(column)

    def lowLeftSearchPosFunc(self, tv, search_dialog, user_data):
        alloc = tv.get_allocation()
        window = tv.get_toplevel().get_window()
        x_loc = alloc.x + window.get_position()[0]
        y_loc = alloc.y + window.get_position()[1] + alloc.height
        search_dialog.move(x_loc, y_loc)
        search_dialog.show_all()

    def pixCompareFunction(self, treemodel, iter0, iter1, column):
        pix0 = treemodel.get_value(iter0, column)
        pix1 = treemodel.get_value(iter1, column)
        if isinstance(pix0, GdkPixbuf.Pixbuf) and isinstance(pix1, GdkPixbuf.Pixbuf):
            return cmp(pix0.get_pixels(), pix1.get_pixels())
        return cmp(pix0, pix1)

    def timeCompareFunction(self, treemodel, iter0, iter1, column):
        (minute0, minute1) = (
            treemodel.get_value(iter0, 8),
            treemodel.get_value(iter1, 8),
        )
        return cmp(minute0, minute1)

    def on_accept(self, button):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        sought = model.get_value(sel_iter, 0)
        if isinstance(sought, FICSChallenge):
            self.connection.om.acceptIndex(sought.index)
        else:
            self.connection.om.playIndex(sought.index)

        try:
            message = self.messages[hash(sought)]
        except KeyError:
            pass
        else:
            message.dismiss()
            del self.messages[hash(sought)]

    def on_assess(self, widget):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        sought = model.get_value(sel_iter, 0)
        player1 = self.connection.username
        player2 = sought.player.name
        game_type = sought.game_type.short_fics_name
        self.connection.glm.assess(player1, player2, game_type)
        self.assess_sent = True

    def on_observe(self, widget, *args):
        player = self.getSelectedPlayer()
        if player is not None:
            if player.game is not None:
                self.connection.bm.observe(player.game)
            else:
                self.connection.bm.observe(None, player=player)

    def on_follow(self, widget):
        player = self.getSelectedPlayer()
        if player is not None:
            self.connection.bm.follow(player)

    def on_chat(self, button):
        player = self.getSelectedPlayer()
        if player is None:
            return
        self.lounge.chat.openChatWithPlayer(player.name)
        # TODO: isadmin og type

    def on_challenge(self, widget):
        player = self.getSelectedPlayer()
        if player is not None:
            self.lounge.seek_challenge.onChallengeButtonClicked(widget, player)

    def on_finger(self, widget):
        player = self.getSelectedPlayer()
        if player is not None:
            self.connection.fm.finger(player.name)
            self.lounge.finger_sent = True

    def on_archived(self, widget):
        player = self.getSelectedPlayer()
        if player is not None:
            self.connection.adm.queryAdjournments(player.name)
            self.connection.adm.queryHistory(player.name)
            self.connection.adm.queryJournal(player.name)



"""Pawns Passed Chess"""

from pychess.Utils.const import PAWNSPASSEDCHESS, VARIANTS_OTHER
from pychess.Utils.Board import Board

PAWNSPASSEDSTART = "rnbqkbnr/8/8/PPPPPPPP/pppppppp/8/8/RNBQKBNR w - - 0 1"


class PawnsPassedBoard(Board):
    """:Description: Standard chess game rules, but where the board setup is defined as all the
    white pawns start on the 5th rank and all the black pawns start on the 4th rank
    """

    variant = PAWNSPASSEDCHESS
    __desc__ = _(
        "FICS wild/8a: http://www.freechess.org/Help/HelpFiles/wild.html\n"
        + "White pawns start on 5th rank and black pawns on the 4th rank"
    )
    name = _("Pawns Passed")
    cecp_name = "normal"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_OTHER

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=PAWNSPASSEDSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



from .lutils.lmove import FILE, RANK


def cmp(x, y):
    return (x > y) - (x < y)


class CordFormatException(Exception):
    pass


class Cord:
    def __init__(self, var1, var2=None, color=None):
        """Inits a new highlevel cord object.
        The cord B3 can be inited in the folowing ways:
        Cord(17), Cord("b3"), Cord(1,2), Cord("b",3)
        color can be one of "R", "G", "B", "Y"
        """
        self.color = color
        if var2 is None:
            if isinstance(var1, int):
                # We assume the format Cord(17)
                self.x = FILE(var1)
                self.y = RANK(var1)
            else:
                # We assume the format Cord("b3")
                self.x = self.charToInt(var1[0])
                self.y = int(var1[1]) - 1
        else:
            if isinstance(var1, str):
                # We assume the format Cord("b",3)
                self.x = self.charToInt(var1)
                self.y = var2 - 1
            else:
                # We assume the format Cord(1,2)
                self.x = var1
                self.y = var2

    def _get_cord(self):
        return self.y * 8 + self.x

    cord = property(_get_cord)

    def _get_cx(self):
        return self.intToChar(self.x)

    cx = property(_get_cx)

    def _get_cy(self):
        return str(self.y + 1)

    cy = property(_get_cy)

    def intToChar(self, x):
        # assert 0 <= x <= 7
        return chr(x + ord("a"))

    def charToInt(self, char):
        ord_char = ord(char)
        if ord("A") <= ord_char <= ord("H"):
            ord_char -= ord("A")
        elif ord("a") <= ord_char <= ord("h"):
            ord_char -= ord("a")
        else:
            raise CordFormatException("x < 0 || x > 7 (%s, %d)" % (char, ord_char))
        return ord_char

    def _set_cords(self, x_y):
        self.x, self.y = x_y

    def _get_cords(self):
        return (self.x, self.y)

    cords = property(_get_cords, _set_cords)

    def __cmp__(self, other):
        if other is None:
            return 1
        if cmp(self.x, other.x):
            return cmp(self.x, other.x)
        if cmp(self.y, other.y):
            return cmp(self.y, other.y)
        return 0

    def __eq__(self, other):
        return other is not None and other.x == self.x and other.y == self.y

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return self.cx + self.cy

    def __hash__(self):
        return self.x * 8 + self.y



from time import strftime, localtime
import random

from gi.repository import Gtk, Gdk, Pango, GObject

from pychess.System import uistuff
from pychess.widgets import insert_formatted
from pychess.Utils.IconLoader import load_icon
from pychess.ic.ICGameModel import ICGameModel


class ChatView(Gtk.Box):
    __gsignals__ = {
        "messageAdded": (GObject.SignalFlags.RUN_FIRST, None, (str, str, object)),
        "messageTyped": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self, gamemodel=None):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        self.gamemodel = gamemodel

        # States for the color generator
        self.colors = {}
        self.startpoint = random.random()

        # Inits the read view
        self.readView = Gtk.TextView()

        self.sw = Gtk.ScrolledWindow()
        self.sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.sw.set_shadow_type(Gtk.ShadowType.ETCHED_IN)

        self.sw.add(self.readView)
        self.readView.set_editable(False)
        self.readView.set_cursor_visible(False)
        self.readView.props.wrap_mode = Gtk.WrapMode.WORD
        self.readView.props.pixels_below_lines = 1
        self.readView.props.pixels_above_lines = 2
        self.readView.props.left_margin = 2

        if isinstance(self.gamemodel, ICGameModel):
            self.refresh = Gtk.Image()
            self.refresh.set_from_pixbuf(load_icon(16, "view-refresh", "stock-refresh"))
            label = _("Observers")
            self.obs_btn = Gtk.Button()
            self.obs_btn.set_image(self.refresh)
            self.obs_btn.set_label(label)
            self.obs_btn_cid = self.obs_btn.connect("clicked", self.on_obs_btn_clicked)

            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

            # Inits the observers view
            self.obsView = Gtk.TextView()
            self.obsView.set_cursor_visible(False)

            self.obsView.set_editable(False)
            self.obsView.props.wrap_mode = Gtk.WrapMode.WORD
            self.obsView.props.pixels_below_lines = 1
            self.obsView.props.pixels_above_lines = 2
            self.obsView.props.left_margin = 2

            text_buffer = self.obsView.get_buffer()
            iter = text_buffer.get_end_iter()
            anchor1 = text_buffer.create_child_anchor(iter)
            self.obsView.add_child_at_anchor(self.obs_btn, anchor1)
            self.button_tag = text_buffer.create_tag("observers")
            text_buffer.insert_with_tags_by_name(iter, " ", "observers")
            text_buffer.insert(iter, "\n")

            if not self.gamemodel.offline_lecture:
                vbox.pack_start(self.obsView, False, True, 0)
            vbox.pack_start(self.sw, True, True, 0)

            self.pack_start(vbox, True, True, 0)
        else:
            self.pack_start(self.sw, True, True, 0)

        # Create a 'log mark' in the beginning of the text buffer. Because we
        # query the log asynchronously and in chunks, we can use this to insert
        # it correctly after previous log messages, but before the new messages.
        start = self.readView.get_buffer().get_start_iter()
        self.readView.get_buffer().create_mark("logMark", start)

        # Inits the write view
        self.writeView = Gtk.Entry()

        box = Gtk.Box()
        self.pack_start(self.writeView, False, False, 0)
        box.add(self.writeView)

        if self.gamemodel is not None and self.gamemodel.offline_lecture:
            label = _("Go on")
            self.go_on_btn = Gtk.Button()
            self.go_on_btn.set_label(label)
            self.go_on_btn_cid = self.go_on_btn.connect(
                "clicked", lambda btn: self.gamemodel.lecture_skip_event.set()
            )
            box.add(self.go_on_btn)

            label = _("Pause")
            self.pause_btn = Gtk.Button()
            self.pause_btn.set_label(label)
            self.pause_btn_cid = self.pause_btn.connect(
                "clicked", lambda btn: self.gamemodel.lecture_pause_event.set()
            )
            box.add(self.pause_btn)

        self.pack_start(box, False, False, 0)

        self.writeview_cid = self.writeView.connect("key-press-event", self.onKeyPress)
        self.cid = None
        if self.gamemodel is not None:
            self.cid = self.gamemodel.connect_after(
                "game_terminated", self.on_game_terminated
            )

    def on_game_terminated(self, model):
        if isinstance(self.gamemodel, ICGameModel):
            self.obs_btn.disconnect(self.obs_btn_cid)
        self.writeView.disconnect(self.writeview_cid)
        if self.cid is not None:
            self.gamemodel.disconnect(self.cid)

    def on_obs_btn_clicked(self, other):
        if not self.gamemodel.connection.ICC:
            allob = "allob " + str(self.gamemodel.ficsgame.gameno)
            self.gamemodel.connection.client.run_command(allob)

    def update_observers(self, other, observers):
        """Rebuilds observers list text"""
        text_buf = self.obsView.get_buffer()
        start_iter = text_buf.get_end_iter()
        start_iter.backward_to_tag_toggle(self.button_tag)
        start_iter.forward_char()
        end_iter = text_buf.get_end_iter()
        text_buf.delete(start_iter, end_iter)
        iter = text_buf.get_end_iter()

        obs_list = observers.split()
        for player in obs_list:
            # Colourize only players able to interact with chat View
            if player.endswith("(U)"):
                text_buf.insert(iter, "%s " % player[:-3])
            elif "(" in player:
                pref = player.split("(", 1)[0]
                self._ensureColor(pref)
                text_buf.insert_with_tags_by_name(iter, "%s " % player, pref + "_bold")
            else:
                text_buf.insert(iter, "%s " % player)
        self.obsView.show_all()

    def _ensureColor(self, pref):
        """Ensures that the tags for pref_normal and pref_bold are set in the text buffer"""
        text_buf = self.readView.get_buffer()
        if pref not in self.colors:
            color = uistuff.genColor(len(self.colors) + 1, self.startpoint)
            self.colors[pref] = color
            color = [int(c * 255) for c in color]
            color = "#" + "".join([hex(v)[2:].zfill(2) for v in color])
            text_buf.create_tag(pref + "_normal", foreground=color)
            text_buf.create_tag(
                pref + "_bold", foreground=color, weight=Pango.Weight.BOLD
            )
            if isinstance(self.gamemodel, ICGameModel):
                otb = self.obsView.get_buffer()
                otb.create_tag(pref + "_normal", foreground=color)
                otb.create_tag(
                    pref + "_bold", foreground=color, weight=Pango.Weight.BOLD
                )

    def clear(self):
        self.writeView.get_buffer().props.text = ""
        self.readView.get_buffer().props.text = ""
        tagtable = self.readView.get_buffer().get_tag_table()
        for i in range(len(self.colors)):
            tagtable.remove("%d_normal" % i)
            tagtable.remove("%d_bold" % i)
        self.colors.clear()

    def __addMessage(self, iter, time, sender, text):
        pref = sender.lower()
        text_buffer = self.readView.get_buffer()
        iter = text_buffer.get_end_iter()
        text_buffer.create_mark("end", iter, False)
        if text_buffer.props.text:
            text_buffer.insert(iter, "\n")

        # Calculate a color for the sender
        self._ensureColor(pref)
        # Insert time, name and text with different stylesd
        text_buffer.insert_with_tags_by_name(iter, "(%s) " % time, pref + "_normal")
        text_buffer.insert_with_tags_by_name(iter, sender + ": ", pref + "_bold")
        insert_formatted(self.readView, iter, text)

        # Scroll the mark onscreen.
        mark = text_buffer.get_mark("end")
        text_buffer.move_mark(mark, iter)
        self.readView.scroll_mark_onscreen(mark)

        # This is used to buzz the user and add senders to a list of active participants
        self.emit("messageAdded", sender, text, self.colors[pref])

    def insertLogMessage(self, timestamp, sender, text):
        """Takes a list of (timestamp, sender, text) pairs, and inserts them in
        the beginning of the document.
        All text will be in a gray color
        """
        text_buffer = self.readView.get_buffer()
        iter = text_buffer.get_iter_at_mark(text_buffer.get_mark("logMark"))
        time = strftime("%H:%M:%S", localtime(timestamp))
        self.__addMessage(iter, time, sender, text)

    def addMessage(self, sender, text):
        text_buffer = self.readView.get_buffer()
        iter = text_buffer.get_end_iter()
        self.__addMessage(iter, strftime("%H:%M:%S"), sender, text)

    def disable(self, message):
        """Sets the write field insensitive, in cases where the channel is
        read only. Use the message to give the user a propriate
        exlpanation"""
        self.writeView.set_sensitive(False)
        self.writeView.set_text(message)

    def enable(self):
        self.writeView.set_text("")
        self.writeView.set_sensitive(True)

    def onKeyPress(self, widget, event):
        if event.keyval in list(map(Gdk.keyval_from_name, ("Return", "KP_Enter"))):
            if not event.get_state() & Gdk.ModifierType.CONTROL_MASK:
                buffer = self.writeView.get_buffer()
                if buffer.props.text:
                    self.emit("messageTyped", buffer.props.text)
                    buffer.props.text = ""
                return True



"""Sets the application background image"""

from os import path

from gi.repository import Gtk, Gdk, GdkPixbuf
import cairo

from pychess.System import conf, uistuff
from pychess.System.prefix import addDataPrefix, addUserCachePrefix

surface = None
provider = None
loldcolor = None
doldcolor = None


def hexcol(color):
    """Description : Takes a colour tuple(rgb) and returns a hex based string #rrggbb
    Returns : (str)
    """
    return "#{:02X}{:02X}{:02X}".format(
        int(color.red * 255), int(color.green * 255), int(color.blue * 255)
    )


def giveBackground(widget):
    widget.connect("draw", expose)
    widget.connect("style-updated", newTheme)


def expose(widget, context):
    cairo_create = widget.get_window().cairo_create()

    x_loc = widget.get_allocation().x
    y_loc = widget.get_allocation().y
    width = widget.get_allocation().width
    height = widget.get_allocation().height
    cairo_create.rectangle(x_loc, y_loc, width, height)
    if not surface:
        newTheme(widget)
    cairo_create.set_source_surface(surface, 0, 0)
    pattern = cairo_create.get_source()
    pattern.set_extend(cairo.EXTEND_REPEAT)
    cairo_create.fill()


def newTheme(widget, background=None):
    global surface, provider, loldcolor, doldcolor

    style_ctxt = widget.get_style_context()

    # get colors from theme

    # bg color
    found, bgcol = style_ctxt.lookup_color("bg_color")
    if not found:
        found, bgcol = style_ctxt.lookup_color("theme_bg_color")
        if not found:
            # fallback value
            bgcol = Gdk.RGBA(red=0.929412, green=0.929412, blue=0.929412, alpha=1.0)

    # bg selected color
    found, bgsel = style_ctxt.lookup_color("theme_selected_bg_color")
    if not found:
        # fallback value
        bgsel = Gdk.RGBA(red=0.290, green=0.565, blue=0.851, alpha=1.0)

    # fg color
    found, fgcol = style_ctxt.lookup_color("fg_color")
    if not found:
        found, fgcol = style_ctxt.lookup_color("theme_fg_color")
        if not found:
            fgcol = Gdk.RGBA(
                red=0.180392, green=0.203922, blue=0.211765, alpha=1.000000
            )

    # base color
    found, basecol = style_ctxt.lookup_color("base_color")
    if not found:
        found, basecol = style_ctxt.lookup_color("theme_base_color")
        if not found:
            basecol = Gdk.RGBA(red=0.929412, green=0.929412, blue=0.929412, alpha=1.0)

    # text color
    found, textcol = style_ctxt.lookup_color("text_color")
    if not found:
        found, textcol = style_ctxt.lookup_color("theme_text_color")
        if not found:
            textcol = Gdk.RGBA(red=0.180392, green=0.203922, blue=0.211765, alpha=1.0)

    def get_col(col, mult):
        red = col.red * mult
        green = col.green * mult
        blue = col.blue * mult
        if red > 1.0:
            red = 1.0
        if green > 1.0:
            green = 1.0
        if blue > 1.0:
            blue = 1.0
        if red == 1 and green == 1 and blue == 1:
            return Gdk.RGBA(0.99, 0.99, 0.99, 1.0)
        else:
            return Gdk.RGBA(red, green, blue, 1.0)

    # derive other colors
    bgacol = get_col(bgcol, 0.9)  # bg_active
    dcol = get_col(bgcol, 0.7)  # dark
    darksel = get_col(bgsel, 0.71)  # dark selected
    dpcol = get_col(bgcol, 0.71)  # dark prelight
    dacol = get_col(dcol, 0.9)  # dark_active
    lcol = get_col(bgcol, 1.3)  # light color
    lightsel = get_col(bgsel, 1.3)  # light selected
    fgsel = Gdk.RGBA(0.99, 0.99, 0.99, 1.0)  # fg selected
    fgpcol = get_col(fgcol, 1.054)  # fg prelight
    fgacol = Gdk.RGBA(0.0, 0.0, 0.0, 1.0)  # fg active
    textaacol = Gdk.RGBA(
        min((basecol.red + textcol.red) / 2.0, 1.0),
        min((basecol.green + textcol.green) / 2.0, 1.0),
        min((basecol.blue + textcol.blue) / 2.0, 1.0),
    )  # text_aa

    data = (
        "@define-color p_bg_color " + hexcol(bgcol) + ";"
        "@define-color p_bg_prelight " + hexcol(bgcol) + ";"
        "@define-color p_bg_active " + hexcol(bgacol) + ";"
        "@define-color p_bg_selected " + hexcol(bgsel) + ";"
        "@define-color p_bg_insensitive " + hexcol(bgcol) + ";"
        "@define-color p_base_color " + hexcol(basecol) + ";"
        "@define-color p_dark_color " + hexcol(dcol) + ";"
        "@define-color p_dark_prelight " + hexcol(dpcol) + ";"
        "@define-color p_dark_active " + hexcol(dacol) + ";"
        "@define-color p_dark_selected " + hexcol(darksel) + ";"
        "@define-color p_text " + hexcol(textcol) + ";"
        "@define-color p_text_aa " + hexcol(textaacol) + ";"
        "@define-color p_light_color " + hexcol(lcol) + ";"
        "@define-color p_light_selected " + hexcol(lightsel) + ";"
        "@define-color p_fg_color " + hexcol(fgcol) + ";"
        "@define-color p_fg_prelight " + hexcol(fgpcol) + ";"
        "@define-color p_fg_selected " + hexcol(fgsel) + ";"
        "@define-color p_fg_active " + hexcol(fgacol) + ";"
    )

    if provider is not None:
        style_ctxt.remove_provider_for_screen(Gdk.Screen.get_default(), provider)

    provider = Gtk.CssProvider.new()
    provider.load_from_data(data.encode())
    style_ctxt.add_provider_for_screen(
        Gdk.Screen.get_default(), provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    )

    lnewcolor = bgcol
    dnewcolor = dcol

    # check if changed
    if loldcolor and background is None:
        if (
            lnewcolor.red == loldcolor.red
            and lnewcolor.green == loldcolor.green
            and lnewcolor.blue == loldcolor.blue
            and dnewcolor.red == doldcolor.red
            and dnewcolor.green == doldcolor.green
            and dnewcolor.blue == doldcolor.blue
        ):
            return

    loldcolor = lnewcolor
    doldcolor = dnewcolor

    # global colors have been set up
    # now set colors on startup panel
    lnewcolor = style_ctxt.lookup_color("p_bg_color")[1]
    dnewcolor = style_ctxt.lookup_color("p_dark_color")[1]

    colors = [
        int(lnewcolor.red * 255),
        int(lnewcolor.green * 255),
        int(lnewcolor.blue * 255),
        int(dnewcolor.red * 255),
        int(dnewcolor.green * 255),
        int(dnewcolor.blue * 255),
    ]

    if background is None:
        background = conf.get("welcome_image")

    if not path.isfile(background):
        background = addDataPrefix("glade/clear.png")
        conf.set("welcome_image", background)

    if not background.endswith("clear.png"):
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(background)
        x, y, height, width = uistuff.getMonitorBounds()
        pixbuf = pixbuf.scale_simple(height, width, GdkPixbuf.InterpType.BILINEAR)
        # for frmat in GdkPixbuf.Pixbuf.get_formats():
        #     print(frmat.get_extensions())
        surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, 0, None)
        return

    # Check if a cache has been saved
    temppng = addUserCachePrefix("temp.png")
    if path.isfile(temppng):
        with open(temppng, "rb") as fyle:
            # Check if the cache was made while using the same theme
            if list(fyle.read(6)) == colors:
                surface = cairo.ImageSurface.create_from_png(fyle)
                return

    # Get mostly transparant shadowy image
    imgsurface = cairo.ImageSurface.create_from_png(background)
    avgalpha = 108 / 255.0

    surface = cairo.ImageSurface(
        cairo.FORMAT_RGB24, imgsurface.get_width(), imgsurface.get_height()
    )
    ctx = cairo.Context(surface)
    if lnewcolor.blue * 65535 - dnewcolor.blue * 65535 > 0:
        midtone = (
            dnewcolor.red
            * 65535
            / (3 * (lnewcolor.blue * 65535 - dnewcolor.blue * 65535) * (1 - avgalpha))
        )
        ctx.set_source_rgb(
            lnewcolor.red / 2 + dnewcolor.red * midtone / 2,
            lnewcolor.green / 2 + dnewcolor.green * midtone / 2,
            lnewcolor.blue / 2 + dnewcolor.blue * midtone / 2,
        )
        ctx.paint()
    ctx.set_source_surface(imgsurface, 0, 0)
    ctx.paint_with_alpha(0.8)

    # Save a cache for later use. Save 'newcolor' in the frist three pixels
    # to check for theme changes between two instances
    with open(temppng, "wb") as fyle:
        fyle.write(bytes(colors))
        surface.write_to_png(fyle)


def isDarkTheme(widget):
    color = widget.get_style_context().get_background_color(Gtk.StateFlags.NORMAL)
    minimal = min(color.red, color.green, color.blue)
    maximal = max(color.red, color.green, color.blue)
    lightness = (minimal + maximal) / 2
    return lightness < 0.5



from gi.repository import Gtk, GObject


firstRun = True


def run(widgets):
    global firstRun
    if firstRun:
        initialize(widgets)
        firstRun = False
    widgets["player_info"].show_all()


def initialize(widgets):
    def addColumns(treeview, *columns):
        model = Gtk.ListStore(*((str,) * len(columns)))
        treeview.set_model(model)
        treeview.get_selection().set_mode(Gtk.SelectionMode.NONE)
        for i, name in enumerate(columns):
            crt = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn(name, crt, text=i)
            treeview.append_column(column)

    addColumns(widgets["results_view"], "", "Games", "Won", "Drawn", "Lost", "Score")
    model = widgets["results_view"].get_model()
    model.append(("White", "67", "28", "24", "15", "59%"))
    model.append(("Black", "66", "26", "23", "17", "56%"))
    model.append(("Total", "133", "54", "47", "32", "58%"))

    addColumns(
        widgets["rating_view"], "Current", "Initial", "Lowest", "Highest", "Average"
    )
    model = widgets["rating_view"].get_model()
    model.append(("1771", "1734", "1659", "1791", "1700"))

    widgets["history_view"].set_model(Gtk.ListStore(object))
    widgets["history_view"].get_selection().set_mode(Gtk.SelectionMode.NONE)
    widgets["history_view"].append_column(
        Gtk.TreeViewColumn("Player Rating History", HistoryCellRenderer(), data=0)
    )
    widgets["history_view"].get_model().append((1,))

    def hide_window(button, *args):
        widgets["player_info"].hide()
        return True

    widgets["player_info"].connect("delete-event", hide_window)
    widgets["player_info_close_button"].connect("clicked", hide_window)


class HistoryCellRenderer(Gtk.CellRenderer):
    __gproperties__ = {
        "data": (
            GObject.TYPE_PYOBJECT,
            "Data",
            "Data",
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        ),
    }

    def __init__(self):
        self.__gobject_init__()
        self.data = None

    def do_set_property(self, pspec, value):
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
        return getattr(self, pspec.name)

    def on_render(self, window, widget, background_area, rect, expose_area, flags):
        if not self.data:
            return
        cairo = window.cairo_create()
        x_loc, y_loc, width, height = rect.x, rect.y, rect.width, rect.height
        cairo.rectangle(x_loc + 1, y_loc + 1, x_loc + width - 2, y_loc + height - 2)
        cairo.set_source_rgb(0.45, 0.45, 0.45)
        cairo.stroke()

    def on_get_size(self, widget, cell_area=None):
        return (0, 0, -1, 130)



import re
from gi.repository import Gtk, GObject
from pychess.widgets.InfoPanel import Panel

TYPE_PERSONAL, TYPE_CHANNEL, TYPE_GUEST, TYPE_ADMIN, TYPE_COMP, TYPE_BLINDFOLD = range(
    6
)


def get_playername(playername):
    re_m = re.match(r"(\w+)\W*", playername)
    return re_m.groups()[0]


class ViewsPanel(Gtk.Notebook, Panel):
    """:Description: This panel is used to display the main chat text for each of the channel or
    private communication
    """

    __gsignals__ = {
        "channel_content_Changed": (GObject.SignalFlags.RUN_FIRST, None, (str, int))
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.set_show_tabs(False)
        self.set_show_border(False)
        self.id2Widget = {}
        self.connection = connection

        label = Gtk.Label()
        label.set_markup("<big>%s</big>" % _("You have opened no conversations yet"))
        label.props.xalign = 0.5
        label.props.yalign = 0.381966011
        label.props.justify = Gtk.Justification.CENTER
        label.props.wrap = True
        label.props.width_request = 300
        self.append_page(label, None)

        # When a person addresses us directly, ChannelsPanel will emit an
        # additem event and we add a new view. This however means that the first
        # message the user sends isn't registred by our privateMessage handler.
        # Therefore we save all messages sent by this hook, and when later we
        # add new items, we test if anything was already received
        self.messageBuffer = {}

        def globalPersonalMessage(cm, name, title, isadmin, text):
            if name not in self.messageBuffer:
                self.messageBuffer[name] = []
            self.messageBuffer[name].append((title, isadmin, text))

        self.connection.cm.connect("privateMessage", globalPersonalMessage)

    def addItem(self, grp_id, name, grp_type, chat_view):
        chat_view.connect("messageTyped", self.onMessageTyped, grp_id, name, grp_type)
        self.connection.cm.connect(
            "channelMessage", self.onChannelMessage, grp_id, chat_view
        )
        self.connection.cm.connect(
            "privateMessage", self.onPersonMessage, get_playername(name), chat_view
        )

        if grp_type == TYPE_CHANNEL:
            self.connection.cm.connect(
                "channelLog", self.onChannelLog, grp_id, chat_view
            )
            self.connection.cm.getChannelLog(grp_id)
            if not self.connection.cm.mayTellChannel(grp_id):
                chat_view.disable(_("Only registered users may talk to this channel"))

        elif grp_type in (
            TYPE_PERSONAL,
            TYPE_COMP,
            TYPE_GUEST,
            TYPE_ADMIN,
            TYPE_BLINDFOLD,
        ):
            if name in self.messageBuffer:
                for title, isadmin, messagetext in self.messageBuffer[name]:
                    chat_view.addMessage(name, messagetext)
                del self.messageBuffer[name]

        self.addPage(chat_view, grp_id)

    def removeItem(self, grp_id):
        self.removePage(grp_id)

    def selectItem(self, grp_id):
        child = self.id2Widget[grp_id]
        self.set_current_page(self.page_num(child))

    def onChannelLog(self, cm, channel, time, handle, text, name_, chat_view):
        if channel.lower() == name_.lower():
            chat_view.insertLogMessage(time, handle, text)

    def onMessageTyped(self, chat_view, text, grp_id, name, grp_type):
        if grp_type == TYPE_CHANNEL:
            self.connection.cm.tellChannel(grp_id, text)
        elif grp_type == TYPE_PERSONAL:
            self.connection.cm.tellPlayer(get_playername(name), text)
        chat_view.addMessage(self.connection.getUsername(), text)

    def onPersonMessage(self, cm, name, title, isadmin, text, name_, chat_view):
        if name.lower() == name_.lower():
            chat_view.addMessage(name, text)
            self.emit("channel_content_Changed", name_, TYPE_PERSONAL)

    def onChannelMessage(
        self, cm, name, isadmin, isme, channel, text, name_, chat_view
    ):
        if channel.lower() == name_.lower() and not isme:
            chat_view.addMessage(name, text)
            self.emit("channel_content_Changed", channel, TYPE_CHANNEL)

    def addPage(self, widget, grp_id):
        self.id2Widget[grp_id] = widget
        self.append_page(widget, None)
        widget.show_all()

    def removePage(self, grp_id):
        child = self.id2Widget.pop(grp_id)
        self.remove_page(self.page_num(child))



import cairo

from gi.repository import Gtk, Gdk, GObject

from pychess.System.prefix import addDataPrefix
from pychess.widgets.BoardView import union
from .BorderBox import BorderBox


class ChainVBox(Gtk.VBox):
    """Inspired by the GIMP chainbutton widget"""

    __gsignals__ = {"clicked": (GObject.SignalFlags.RUN_FIRST, None, ())}

    def __init__(self):
        GObject.GObject.__init__(self)
        chainline = ChainLine(CHAIN_TOP)
        self.pack_start(chainline, True, True, 0)
        self.button = Gtk.Button()
        self.pack_start(self.button, False, True, 0)
        chainline = ChainLine(CHAIN_BOTTOM)
        self.pack_start(chainline, True, True, 0)

        self.image = Gtk.Image()
        self.image.set_from_file(addDataPrefix("glade/stock-vchain-24.png"))
        self.button.set_image(self.image)
        self.button.set_relief(Gtk.ReliefStyle.NONE)
        self.button.set_property("yalign", 0)
        self._active = True
        self.button.connect("clicked", self.onClicked)

    def getActive(self):
        return self._active

    def setActive(self, active):
        assert isinstance(active, bool)
        self._active = active
        if self._active is True:
            self.image.set_from_file(addDataPrefix("glade/stock-vchain-24.png"))
        else:
            self.image.set_from_file(addDataPrefix("glade/stock-vchain-broken-24.png"))

    active = property(getActive, setActive)

    def onClicked(self, button):
        if self._active is False:
            self.image.set_from_file(addDataPrefix("glade/stock-vchain-24.png"))
            self._active = True
        else:
            self.image.set_from_file(addDataPrefix("glade/stock-vchain-broken-24.png"))
            self._active = False
        self.emit("clicked")


CHAIN_TOP, CHAIN_BOTTOM = range(2)
SHORT_LINE = 2
LONG_LINE = 8


class ChainLine(Gtk.Alignment):
    """The ChainLine's are the little right-angle lines above and below the chain
    button that visually connect the ChainButton to the widgets who's values
    are "chained" together by the ChainButton being active"""

    def __init__(self, position):
        GObject.GObject.__init__(self)
        self.position = position
        self.connect_after("size-allocate", self.on_size_allocate)
        self.connect_after("draw", self.on_draw)
        self.set_size_request(10, 10)
        self.lastRectangle = None

    def on_size_allocate(self, widget, requisition):
        if self.get_window():
            allocation = self.get_allocation()
            rect = Gdk.Rectangle()
            rect.x, rect.y, rect.width, rect.height = (
                allocation.x,
                allocation.y,
                allocation.width,
                allocation.height,
            )
            unionrect = (
                union(self.lastRectangle, rect)
                if self.lastRectangle is not None
                else rect
            )
            self.get_window().invalidate_rect(unionrect, True)
            self.get_window().process_updates(True)
            self.lastRectangle = rect

    def on_draw(self, widget, context):
        self.draw(context)
        return False

    ###
    # the original Gtk.Style.paint_polygon() way to draw, like The GIMP does it
    ###
    #    def draw (self, widget, event):
    #        a = self.get_allocation()
    #        print a.x, a.y, a.width, a.height
    #        points = [None, None, None]
    #        points[0] = (a.x + a.width/2 - SHORT_LINE, a.y + a.height/2)
    #        points[1] = (points[0][0] + SHORT_LINE, points[0][1])
    #        points[2] = (points[1][0], self.position == CHAIN_TOP and a.y+a.height-1 or a.y)
    #        if self.position == CHAIN_BOTTOM:
    #            t = points[0]
    #            points[0] = points[2]
    #            points[2] = t
    #        print points
    #        self.points = points
    #
    #        style = widget.get_style()
    #        style.paint_polygon(widget.get_parent_window(),
    #                            Gtk.StateType.NORMAL,
    #                            Gtk.ShadowType.ETCHED_OUT,
    #                            event.area,
    #                            widget,
    #                            "chainbutton",
    #                            points,
    #                            False)

    def __toAHalf(self, number):
        """To draw thin straight lines in cairo that aren't blurry, you have to
        adjust the endpoints by 0.5: http://www.cairographics.org/FAQ/#sharp_lines"""
        return int(number) + 0.5

    def draw(self, context):
        allocation = self.get_allocation()
        x_loc = allocation.x
        y_loc = allocation.y
        width = allocation.width - 1
        height = allocation.height

        context.set_source_rgb(0.2, 0.2, 0.2)
        #        context.rectangle(0, 0, width, height)
        #        context.fill()

        context.move_to(
            self.__toAHalf(x_loc + width / 2.0) - LONG_LINE,
            self.__toAHalf(y_loc + height / 2.0),
        )
        context.line_to(
            self.__toAHalf(x_loc + width / 2.0), self.__toAHalf(y_loc + height / 2.0)
        )
        if self.position == CHAIN_TOP:
            context.line_to(
                self.__toAHalf(x_loc + width / 2.0),
                self.__toAHalf(float(y_loc + height)),
            )
        else:
            context.line_to(
                self.__toAHalf(x_loc + width / 2.0), self.__toAHalf(y_loc + 0.0)
            )
        context.set_line_width(1.0)
        context.set_line_cap(cairo.LINE_CAP_ROUND)
        context.set_line_join(cairo.LINE_JOIN_ROUND)
        context.stroke()

    def __str__(self):
        allocation = self.get_allocation()
        chain_str = "ChainLine({}, {}, {}, {}".format(
            allocation.x, allocation.y, allocation.width, allocation.height
        )
        chain_str += self.position == CHAIN_TOP and ", CHAIN_TOP" or ", CHAIN_BOTTOM"
        return chain_str + ")"


if __name__ == "__main__":
    win = Gtk.Window()
    chainvbox = ChainVBox()
    label = Gtk.Label(label="Locked")
    adjustment = Gtk.Adjustment(value=10, upper=100, lower=0, step_increment=1)
    spinbutton1 = Gtk.SpinButton(adjustment=adjustment)
    adjustment = Gtk.Adjustment(value=0, upper=100, lower=0, step_increment=1)
    spinbutton2 = Gtk.SpinButton(adjustment=adjustment)
    table = Gtk.Table(rows=3, columns=2)
    #    table.attach(label,0,2,0,1)
    #    table.attach(chainvbox,1,2,1,3)
    #    table.attach(spinbutton1,0,1,1,2)
    #    table.attach(spinbutton2,0,1,2,3)
    table.attach(label, 0, 2, 0, 1, xoptions=Gtk.AttachOptions.SHRINK)
    table.attach(chainvbox, 1, 2, 1, 3, xoptions=Gtk.AttachOptions.SHRINK)
    table.attach(spinbutton1, 0, 1, 1, 2, xoptions=Gtk.AttachOptions.SHRINK)
    table.attach(spinbutton2, 0, 1, 2, 3, xoptions=Gtk.AttachOptions.SHRINK)
    table.set_row_spacings(2)

    def onChainBoxClicked(*whatever):
        if chainvbox.active is False:
            label.set_label("Unlocked")
        else:
            label.set_label("Locked")

    chainvbox.connect("clicked", onChainBoxClicked)
    border_box = BorderBox(widget=table)
    win.add(border_box)
    #    win.resize(150,100)
    win.connect("delete-event", Gtk.main_quit)
    win.show_all()
    Gtk.main()



import os
import atexit
import gettext
import sqlite3

from pychess.System import conf
from pychess.System.prefix import addDataPrefix, isInstalled

ECO_MAIN_LANG = "en"
ECO_LANGS = [ECO_MAIN_LANG, "da", "de", "es", "hu", "fr"]

# EN = https://github.com/niklasf/eco
# DA = https://da.wikipedia.org/wiki/Skak%C3%A5bninger
# DE = https://de.wikipedia.org/wiki/ECO-Code
# ES = https://es.wikipedia.org/wiki/Anexo:Aperturas_de_ajedrez
# HU = https://hu.wikipedia.org/wiki/Sakkmegnyit%C3%A1sok_list%C3%A1ja
# FR = https://fr.wikipedia.org/wiki/Liste_des_ouvertures_d'%C3%A9checs_suivant_le_code_ECO

db_path = os.path.join(addDataPrefix("eco.db"))
if os.path.exists(db_path):
    conn = sqlite3.connect(db_path, check_same_thread=False)
    atexit.register(conn.close)
    ECO_OK = True
else:
    print("Warning: eco.db not found, please run pgn2ecodb.sh")
    ECO_OK = False

if isInstalled():
    mofile = gettext.find("pychess")
else:
    mofile = gettext.find("pychess", localedir=addDataPrefix("lang"))

if mofile is None:
    lang = ECO_MAIN_LANG
else:
    lang = mofile.split(os.sep)[-3]


def get_eco(hash, exactPosition=True):
    if not ECO_OK:
        return None
    cur = conn.cursor()
    select = "select eco, opening, variation, endline from openings where hash=? and hkey=? and lang=? and endline like ?"
    qh = hex(hash)[2:]
    qhkey = int(qh[-2:], 16)
    qlang = ECO_MAIN_LANG if conf.no_gettext or lang not in ECO_LANGS else lang
    qpos = 1 if exactPosition else "%"
    cur.execute(select, (qh, qhkey, qlang, qpos))
    result = cur.fetchone()
    if result is None and qlang != ECO_MAIN_LANG:
        cur.execute(select, (qh, qhkey, ECO_MAIN_LANG, qpos))
        result = cur.fetchone()
    return result


def find_opening_fen(keyword):
    # Checks
    if not ECO_OK:
        return None
    if keyword is None:
        return None
    keyword = keyword.strip()
    if keyword == "":
        return None
    cur = conn.cursor()

    # Languages to check
    langs = [ECO_MAIN_LANG]
    lang = os.getenv("LANG", ECO_MAIN_LANG).lower()
    if lang in ECO_LANGS and lang != ECO_MAIN_LANG:
        langs = [lang] + langs

    # Execute the check
    for lang in langs:
        # Lookup by ECO
        query = "select fen from openings   \
                 where mainline = 1          \
                   and endline = 1            \
                   and eco = ?                 \
                   and lang = ?"
        cur.execute(query, (keyword.upper(), lang))
        result = cur.fetchone()
        if result is not None:
            return result[0]

        # Lookup by detailed name
        query = "select fen from openings               \
                 where endline = 1                       \
                   and lang = ?                           \
                   and (  lower(opening) like ?            \
                       or lower(variation) like ?           \
                   )                                         \
                 order by eco, opening, length(variation)     \
                 limit 1"
        kwenh = ("%%%s%%" % keyword.lower()).replace("*", "%")
        cur.execute(query, (lang, kwenh, kwenh))
        result = cur.fetchone()
        if result is not None:
            return result[0]

    # Nothing found
    return None



"""The King of the Hill Variation"""

from pychess.Utils.const import (
    KINGOFTHEHILLCHESS,
    VARIANTS_OTHER_NONSTANDARD,
    D4,
    D5,
    E4,
    E5,
)
from pychess.Utils.Board import Board


class KingOfTheHillBoard(Board):
    """:Description: The King of the hill variation is where the object of the game
    is to try and manoeuvre to the centre of the board. The gmae is won when one player
    manages to get there king to any of the 4 centre square ie d4, d5, e4, e5
    """

    variant = KINGOFTHEHILLCHESS
    __desc__ = _(
        "Bringing your king legally to the center (e4, d4, e5, d5) instantly wins the game!\n"
        + "Normal rules apply in other cases and checkmate also ends the game."
    )
    name = _("King of the hill")
    cecp_name = "kingofthehill"
    need_initial_board = False
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD


def testKingInCenter(board):
    """Test for a winning position"""
    return board.kings[board.color - 1] in (E4, E5, D4, D5)



import asyncio
import os

from gi.repository import Gdk, Gtk, GObject, Pango, PangoCairo

from pychess.System import conf, uistuff
from pychess.Utils import prettyPrintScore
from pychess.Utils.const import (
    HINT,
    OPENING,
    SPY,
    BLACK,
    ENDGAME,
    DRAW,
    WHITEWON,
    WHITE,
    NORMALCHESS,
)
from pychess.Utils.book import getOpenings
from pychess.Utils.eco import get_eco
from pychess.Utils.IconLoader import get_pixbuf
from pychess.Utils.logic import legalMoveCount
from pychess.Utils.Move import Move, toSAN, toFAN, listToMoves
from pychess.Utils.lutils.lmovegen import newMove
from pychess.Utils.lutils.lmove import ParsingError
from pychess.System.prefix import addDataPrefix
from pychess.System.Log import log
from math import ceil

__title__ = _("Hints")

__icon__ = addDataPrefix("glade/panel_book.svg")

__desc__ = _(
    "The hint panel will provide computer advice during each stage of the game"
)

__about__ = _("Official PyChess panel.")


class Advisor:
    def __init__(self, store, name, mode):
        """The tree store's columns are:
        (Board, Move, pv)           Indicate the suggested move
        text or barWidth or goodness  Indicate its strength (last 2 are 0 to 1.0)
        pvlines                     Number of analysis lines for analysing engines
        is pvlines editable         Boolean
        Details                     Describe a PV, opening name, etc.
        star/stop                   Boolean HINT, SPY analyzing toggle button state
        is start/stop visible       Boolean"""

        self.store = store
        self.name = name
        self.mode = mode
        store.append(None, self.textOnlyRow(name, mode))

    @property
    def path(self):
        for i, row in enumerate(self.store):
            if row[4] == self.name:
                return (i,)

    def shownChanged(self, boardview, shown):
        """Update the suggestions to match a changed position."""
        pass

    def gamewidget_closed(self, gamewidget):
        pass

    def child_tooltip(self, i):
        """Return a tooltip (or empty) string for the given child row."""
        return ""

    def row_activated(self, path, model):
        """Act on a double-clicked child row other than a move suggestion."""
        pass

    def query_tooltip(self, path):
        indices = path.get_indices()
        if not indices[1:]:
            return self.tooltip
        return self.child_tooltip(indices[1])

    def empty_parent(self):
        while True:
            parent = self.store.get_iter(self.path)
            child = self.store.iter_children(parent)
            if not child:
                return parent
            self.store.remove(child)

    def textOnlyRow(self, text, mode=None):
        return [
            (None, None, None),
            ("", 0, None),
            0,
            False,
            text,
            False,
            mode in (HINT, SPY),
        ]

    def _del(self):
        pass


class OpeningAdvisor(Advisor):
    def __init__(self, store, tv, boardcontrol):
        Advisor.__init__(self, store, _("Opening Book"), OPENING)
        self.tooltip = _(
            "The opening book will try to inspire you during the opening phase of the game by showing you common moves made by chess masters"
        )
        #        self.opening_names = []
        self.tv = tv
        self.boardcontrol = boardcontrol
        self.boardview = boardcontrol.view

    def shownChanged(self, boardview, shown):
        m = boardview.model
        if m is None or m.isPlayingICSGame():
            return

        b = m.getBoardAtPly(shown, boardview.shown_variation_idx)
        parent = self.empty_parent()

        openings = getOpenings(b.board)
        openings.sort(key=lambda t: t[1], reverse=True)
        if not openings:
            return

        totalWeight = 0.0
        # Polyglot-formatted books have space for learning data.
        # See version ac31dc37ec89 for an attempt to parse it.
        # In this version, we simply ignore it. (Most books don't have it.)
        for move, weight, learn in openings:
            totalWeight += weight

        self.opening_names = []
        for move, weight, learn in openings:
            if totalWeight != 0:
                weight /= totalWeight
            goodness = min(float(weight * len(openings)), 1.0)
            weight = "%0.1f%%" % (100 * weight)

            opening = get_eco(
                b.move(Move(move)).board.hash,
                exactPosition=conf.get("book_exact_match"),
            )
            if opening is None:
                eco = ""
            #                self.opening_names.append("")
            else:
                eco = "{} {}{}{}".format(
                    opening[0], opening[1], ", " if opening[2] != "" else "", opening[2]
                )
                #                self.opening_names.append("%s %s" % (opening[1], opening[2]))
                if opening[3] == int(False):
                    eco = "(...) %s" % eco.strip()

            self.store.append(
                parent,
                [
                    (b, Move(move), None),
                    (weight, 1, goodness),
                    0,
                    False,
                    eco,
                    False,
                    False,
                ],
            )
        tp = Gtk.TreePath(self.path)
        self.tv.expand_row(tp, False)

    #    def child_tooltip (self, i):
    #        return "" if len(self.opening_names)==0 else self.opening_names[i]

    def row_activated(self, iter, model, from_gui=True):
        if self.store.get_path(iter) != Gtk.TreePath(self.path):
            board, move, moves = self.store[iter][0]
            self.boardcontrol.play_or_add_move(board, move)


class EngineAdvisor(Advisor):
    # An EngineAdvisor always has self.linesExpected rows reserved for analysis.
    def __init__(self, store, engine, mode, tv, boardcontrol):
        if mode == HINT:
            Advisor.__init__(self, store, _("Analysis by %s") % engine, HINT)
            self.tooltip = (
                _(
                    "%s will try to predict which move is best and which side has the advantage"
                )
                % engine
            )
        else:
            Advisor.__init__(self, store, _("Threat analysis by %s") % engine, SPY)
            self.tooltip = (
                _(
                    "%s will identify what threats would exist if it were your opponent's turn to move"
                )
                % engine
            )
        self.engine = engine
        self.tv = tv
        self.active = False
        self.linesExpected = 1
        self.boardview = boardcontrol.view

        self.cid1 = self.engine.connect("analyze", self.on_analyze)
        self.cid2 = self.engine.connect("readyForOptions", self.on_ready_for_options)
        self.cid3 = self.engine.connect_after("readyForMoves", self.on_ready_for_moves)

        self.figuresInNotation = conf.get("figuresInNotation")

        def on_figures_in_notation(none):
            self.figuresInNotation = conf.get("figuresInNotation")

        self.cid4 = conf.notify_add("figuresInNotation", on_figures_in_notation)

    def _del(self):
        self.engine.disconnect(self.cid1)
        self.engine.disconnect(self.cid2)
        self.engine.disconnect(self.cid3)
        conf.notify_remove(self.cid4)

    def _create_new_expected_lines(self):
        parent = self.empty_parent()
        for line in range(self.linesExpected):
            self.store.append(parent, self.textOnlyRow(_("Calculating...")))
        self.tv.expand_row(Gtk.TreePath(self.path), False)
        return parent

    def shownChanged(self, boardview, shown):
        m = boardview.model
        if m is None:
            return
        if m.isPlayingICSGame() and not m.lesson_game:
            return

        self.engine.setBoard(
            boardview.model.getBoardAtPly(shown, boardview.shown_variation_idx),
            search=self.active or m.lesson_game,
        )

        if self.active:
            self._create_new_expected_lines()

    def on_ready_for_options(self, engine):
        engine_max = self.engine.maxAnalysisLines()
        engine_value = self.engine.getAnalysisLines()
        self.linesExpected = engine_value if engine_value <= engine_max else engine_max

        m = self.boardview.model
        if m.isPlayingICSGame():
            return

        parent = self._create_new_expected_lines()

        # set pvlines, but set it 0 if engine max is only 1
        self.store.set_value(parent, 2, 0 if engine_max == 1 else self.linesExpected)
        # set it editable
        self.store.set_value(parent, 3, engine_max > 1)
        # set start/stop cb visible
        self.store.set_value(parent, 6, True)
        self.active = True

    def on_ready_for_moves(self, engine):
        self.shownChanged(self.boardview, self.boardview.shown)

    def on_analyze(self, engine, analysis):
        if self.boardview.animating:
            return

        if self.boardview.model.isPlayingICSGame():
            return

        if not self.active:
            return

        for i, line in enumerate(analysis):
            if line is None:
                self.store[self.path + (i,)] = self.textOnlyRow("")
                continue

            ply, movstrs, score, depth, nps = line
            board0 = self.engine.board
            board = board0.clone()
            try:
                pv = listToMoves(board, movstrs, validate=True)
            except ParsingError as e:
                # ParsingErrors may happen when parsing "old" lines from
                # analyzing engines, which haven't yet noticed their new tasks
                log.debug(
                    "EngineAdvisor.on_analyze(): Ignored (%s) from analyzer: ParsingError%s"
                    % (" ".join(movstrs), e)
                )
                return

            move = None
            if pv:
                move = pv[0]

            ply0 = board.ply if self.mode == HINT else board.ply + 1
            counted_pv = []
            for j, pvmove in enumerate(pv):
                ply = ply0 + j
                if ply % 2 == 0:
                    mvcount = "%d." % (ply / 2 + 1)
                elif j == 0:
                    mvcount = "%d..." % (ply / 2 + 1)
                else:
                    mvcount = ""
                counted_pv.append(
                    "%s%s"
                    % (
                        mvcount,
                        (
                            toFAN(board, pvmove)
                            if self.figuresInNotation
                            else toSAN(board, pvmove, True)
                        ),
                    )
                )
                board = board.move(pvmove)

            goodness = (min(max(score, -250), 250) + 250) / 500.0
            if self.engine.board.color == BLACK:
                score = -score

            self.store[self.path + (i,)] = [
                (board0, move, pv),
                (prettyPrintScore(score, depth, format_mate=True), 1, goodness),
                0,
                False,
                " ".join(counted_pv),
                False,
                False,
            ]

    def start_stop(self, tb):
        if not tb:
            self.active = True
            self.boardview.model.resume_analyzer(self.mode)
        else:
            self.active = False
            self.boardview.model.pause_analyzer(self.mode)

    def multipv_edited(self, value):
        value = self.engine.requestMultiPV(value)
        if value != self.linesExpected:
            parent = self.store.get_iter(self.path)
            if value > self.linesExpected:
                while self.linesExpected < value:
                    self.store.append(parent, self.textOnlyRow(_("Calculating...")))
                    self.linesExpected += 1
            else:
                while self.linesExpected > value:
                    child = self.store.iter_children(parent)
                    if child is not None:
                        self.store.remove(child)
                    self.linesExpected -= 1
        return value

    def row_activated(self, iter, model):
        if not self.active:
            return

        if self.mode == HINT and self.store.get_path(iter) != Gtk.TreePath(self.path):
            moves = self.store[iter][0][2]
            if moves is not None:
                # score = self.store[iter][1][0]
                model.add_variation(self.engine.board, moves)

        if self.mode == SPY and self.store.get_path(iter) != Gtk.TreePath(self.path):
            moves = self.store[iter][0][2]
            if moves is not None:
                # score = self.store[iter][1][0]
                board = self.engine.board.board
                # SPY analyzer has inverted color boards
                # we need to chage it to get the board in gamemodel variations board list later
                board.setColor(1 - board.color)
                king = board.kings[board.color]
                null_move = Move(newMove(king, king))
                model.add_variation(self.engine.board, [null_move] + moves)

    def child_tooltip(self, i):
        if self.active:
            if i < self.linesExpected:
                return _(
                    "Engine scores are in units of pawns, from White's point of view. Double clicking on analysis lines you can insert them into Annotation panel as variations."
                )
            else:
                return _(
                    "Adding suggestions can help you find ideas, but slows down the computer's analysis."
                )
        return ""


class EndgameAdvisor(Advisor):
    def __init__(self, store, tv, boardcontrol):
        Advisor.__init__(self, store, _("Endgame Table"), ENDGAME)
        # deferred import to not slow down PyChess starting up
        from pychess.Utils.EndgameTable import EndgameTable

        self.egtb = EndgameTable()
        # If mate in # was activated by double click let egtb do the rest
        self.auto_activate = False
        self.tv = tv
        self.boardcontrol = boardcontrol
        self.boardview = boardcontrol.view
        self.tooltip = _(
            "The endgame table will show exact analysis when there are few pieces on the board."
        )
        # TODO: Show a message if tablebases for the position exist but are neither installed nor allowed.

        self.queue = asyncio.Queue()
        self.egtb_task = asyncio.create_task(self.start())

    class StopNow(Exception):
        pass

    async def start(self):
        while True:
            v = await self.queue.get()
            if isinstance(v, Exception) and v == self.StopNow:
                break
            elif v == self.board.board:
                ret = await self.egtb.scoreAllMoves(v)
                self.on_scored(v, ret)
            self.queue.task_done()

    def shownChanged(self, boardview, shown):
        m = boardview.model
        if m is None or m.variant.variant != NORMALCHESS or m.isPlayingICSGame():
            if not (m.practice_game or m.lesson_game):
                return

        self.parent = self.empty_parent()
        self.board = m.getBoardAtPly(shown, boardview.shown_variation_idx)
        self.queue.put_nowait(self.board.board)

    def _del(self):
        try:
            self.queue.put_nowait(self.StopNow)
        except asyncio.QueueFull:
            log.warning("EndgameAdvisor.gamewidget_closed: Queue.Full")
        self.egtb_task.cancel()

    def on_scored(self, board, endings):
        m = self.boardview.model

        if board != self.board.board:
            return

        for move, result, depth in endings:
            if result == DRAW:
                result = (_("Draw"), 1, 0.5)
                details = ""
            elif (result == WHITEWON) ^ (self.board.color == WHITE):
                result = (_("Loss"), 1, 0.0)
                details = _("Mate in %d") % depth
            else:
                result = (_("Win"), 1, 1.0)
                details = _("Mate in %d") % depth

            if m.practice_game or m.lesson_game:
                m.hint = "{} {} {}".format(
                    toSAN(self.board, move, True), result[0], details
                )
                return

            if m.isPlayingICSGame():
                return

            self.store.append(
                self.parent,
                [(self.board, move, None), result, 0, False, details, False, False],
            )

        self.tv.expand_row(Gtk.TreePath(self.path), False)

        if self.auto_activate:
            path = None
            for i, row in enumerate(self.store):
                if row[4] == self.name:
                    path = Gtk.TreePath.new_from_indices((i, 0))
                    break
            if path is not None:
                self.row_activated(
                    self.tv.get_model().get_iter(path), m, from_gui=False
                )

    def row_activated(self, iter, model, from_gui=True):
        if self.store.get_path(iter) != Gtk.TreePath(self.path):
            board, move, moves = self.store[iter][0]

            if from_gui:
                result = self.store[iter][1]
                if result is not None and result[2] != 0.5:
                    # double click on mate in #
                    self.auto_activate = True

            self.boardcontrol.play_or_add_move(board, move)


class Sidepanel:
    def load(self, gmwidg):
        self.gmwidg = gmwidg
        self.boardcontrol = gmwidg.board
        self.boardview = gmwidg.board.view

        self.figuresInNotation = conf.get("figuresInNotation")

        self.sw = Gtk.ScrolledWindow()
        self.tv = Gtk.TreeView()
        self.tv.set_property("headers_visible", False)
        self.sw.add(self.tv)
        self.sw.show_all()

        self.store = Gtk.TreeStore(
            GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT, int, bool, str, bool, bool
        )
        self.tv.set_model(self.store)

        # ## move suggested
        moveRenderer = Gtk.CellRendererText()
        moveRenderer.set_property("xalign", 1.0)
        moveRenderer.set_property("yalign", 0)
        c0 = Gtk.TreeViewColumn("Move", moveRenderer)

        def getMoveText(column, cell, store, iter, data):
            board, move, pv = store[iter][0]
            if not move:
                cell.set_property("text", "")
            else:
                if self.figuresInNotation:
                    cell.set_property("text", toFAN(board, move))
                else:
                    cell.set_property("text", toSAN(board, move, True))

        c0.set_cell_data_func(moveRenderer, getMoveText)

        # ## strength of the move
        c1 = Gtk.TreeViewColumn("Strength", StrengthCellRenderer(), data=1)

        # ## multipv (number of analysis lines)
        self.multipvRenderer = Gtk.CellRendererSpin()
        adjustment = Gtk.Adjustment(value=1, lower=1, upper=9, step_increment=1)
        self.multipvRenderer.set_property("adjustment", adjustment)
        self.multipvRenderer.set_property("editable", True)
        self.multipvRenderer.set_property("width_chars", 1)
        c2 = Gtk.TreeViewColumn("PV", self.multipvRenderer, editable=3)
        c2.set_property("min_width", 80)

        def spin_visible(column, cell, store, iter, data):
            if store[iter][2] == 0:
                cell.set_property("visible", False)
            else:
                cell.set_property("text", str(store[iter][2]))
                cell.set_property("visible", True)

        c2.set_cell_data_func(self.multipvRenderer, spin_visible)

        def multipv_edited(renderer, path, text):
            iter = self.store.get_iter(path)
            self.store.set_value(
                iter,
                2,
                self.advisors[int(path[0])].multipv_edited(
                    1 if text == "" else int(text)
                ),
            )

        self.multipv_cid = self.multipvRenderer.connect("edited", multipv_edited)

        # ## start/stop button for analysis engines
        self.toggleRenderer = CellRendererPixbufXt()
        self.toggleRenderer.set_property("stock-id", "gtk-add")
        c4 = Gtk.TreeViewColumn("StartStop", self.toggleRenderer)

        self.play_image = get_pixbuf("glade/play.png")
        self.pause_image = get_pixbuf("glade/pause.png")

        def cb_visible(column, cell, store, iter, data):
            if not store[iter][6]:
                cell.set_property("visible", False)
            else:
                cell.set_property("visible", True)

            if store[iter][5]:
                cell.set_property("pixbuf", self.play_image)
            else:
                cell.set_property("pixbuf", self.pause_image)

        c4.set_cell_data_func(self.toggleRenderer, cb_visible)

        def toggled_cb(cell, path):
            self.store[path][5] = not self.store[path][5]
            self.advisors[int(path[0])].start_stop(self.store[path][5])

        self.toggle_cid = self.toggleRenderer.connect("clicked", toggled_cb)

        self.tv.append_column(c4)
        self.tv.append_column(c0)
        self.tv.append_column(c1)
        self.tv.append_column(c2)
        # ## header text, or analysis line
        uistuff.appendAutowrapColumn(self.tv, "Details", text=4)

        self.cid = self.boardview.connect("shownChanged", self.shownChanged)
        self.tv_cids = [
            self.tv.connect("cursor_changed", self.selection_changed),
            self.tv.connect("select_cursor_row", self.selection_changed),
            self.tv.connect("row-activated", self.row_activated),
            self.tv.connect("query-tooltip", self.query_tooltip),
        ]
        self.tv.props.has_tooltip = True
        self.tv.set_property("show-expanders", False)

        self.advisors = []
        self.conf_conids = []

        if conf.get("opening_check"):
            advisor = OpeningAdvisor(self.store, self.tv, self.boardcontrol)
            self.advisors.append(advisor)
        if conf.get("endgame_check"):
            advisor = EndgameAdvisor(self.store, self.tv, self.boardcontrol)
            self.advisors.append(advisor)

        self.model_cids = [
            gmwidg.gamemodel.connect("analyzer_added", self.on_analyzer_added),
            gmwidg.gamemodel.connect("analyzer_removed", self.on_analyzer_removed),
            gmwidg.gamemodel.connect_after("game_terminated", self.on_game_terminated),
        ]

        def on_opening_check(none):
            if conf.get("opening_check") and self.boardview is not None:
                advisor = OpeningAdvisor(self.store, self.tv, self.boardcontrol)
                self.advisors.append(advisor)
                advisor.shownChanged(self.boardview, self.boardview.shown)
            else:
                for advisor in self.advisors:
                    if advisor.mode == OPENING:
                        parent = advisor.empty_parent()
                        self.store.remove(parent)
                        self.advisors.remove(advisor)

        self.conf_conids.append(conf.notify_add("opening_check", on_opening_check))

        def on_opening_file_entry_changed(none):
            path = conf.get("opening_file_entry")
            if os.path.isfile(path):
                for advisor in self.advisors:
                    if advisor.mode == OPENING and self.boardview is not None:
                        advisor.shownChanged(self.boardview, self.boardview.shown)

        self.conf_conids.append(
            conf.notify_add("opening_file_entry", on_opening_file_entry_changed)
        )

        def on_endgame_check(none):
            if conf.get("endgame_check"):
                advisor = EndgameAdvisor(self.store, self.tv, self.boardcontrol)
                self.advisors.append(advisor)
                advisor.shownChanged(self.boardview, self.boardview.shown)
            else:
                for advisor in self.advisors:
                    if advisor.mode == ENDGAME:
                        advisor._del()
                        parent = advisor.empty_parent()
                        self.store.remove(parent)
                        self.advisors.remove(advisor)

        self.conf_conids.append(conf.notify_add("endgame_check", on_endgame_check))

        def on_figures_in_notation(none):
            self.figuresInNotation = conf.get("figuresInNotation")

        self.conf_conids.append(
            conf.notify_add("figuresInNotation", on_figures_in_notation)
        )

        return self.sw

    def on_game_terminated(self, model):
        self.multipvRenderer.disconnect(self.multipv_cid)
        self.toggleRenderer.disconnect(self.toggle_cid)
        for cid in self.tv_cids:
            self.tv.disconnect(cid)
        for conid in self.conf_conids:
            conf.notify_remove(conid)
        for cid in self.model_cids:
            self.gmwidg.gamemodel.disconnect(cid)
        self.boardview.disconnect(self.cid)
        for advisor in self.advisors:
            advisor._del()

    def on_analyzer_added(self, gamemodel, analyzer, analyzer_type):
        if analyzer_type == HINT:
            self.advisors.append(
                EngineAdvisor(self.store, analyzer, HINT, self.tv, self.boardcontrol)
            )
        if analyzer_type == SPY:
            self.advisors.append(
                EngineAdvisor(self.store, analyzer, SPY, self.tv, self.boardcontrol)
            )

    def on_analyzer_removed(self, gamemodel, analyzer, analyzer_type):
        for advisor in self.advisors:
            if advisor.mode == analyzer_type:
                advisor.active = False
                advisor.engine = None
                parent = advisor.empty_parent()
                self.store.remove(parent)
                self.advisors.remove(advisor)

    def shownChanged(self, boardview, shown):
        if boardview.model is None:
            return
        boardview.bluearrow = None

        if (
            legalMoveCount(
                boardview.model.getBoardAtPly(shown, boardview.shown_variation_idx)
            )
            == 0
        ):
            if self.sw.get_child() == self.tv:
                self.sw.remove(self.tv)
                label = Gtk.Label(_("In this position,\nthere is no legal move."))
                label.set_property("yalign", 0.1)
                self.sw.add_with_viewport(label)
                self.sw.get_child().set_shadow_type(Gtk.ShadowType.NONE)
                self.sw.show_all()

                # stop egtb auto activation on mating lines
                for advisor in self.advisors:
                    advisor.auto_activate = False
            return

        for advisor in self.advisors:
            advisor.shownChanged(boardview, shown)
        self.tv.expand_all()

        if self.sw.get_child() != self.tv:
            log.warning("bookPanel.Sidepanel.shownChanged: get_child() != tv")
            self.sw.remove(self.sw.get_child())
            self.sw.add(self.tv)

    def selection_changed(self, widget, *args):
        iter = self.tv.get_selection().get_selected()[1]
        if iter:
            board, move, pv = self.store[iter][0]
            if move is not None:
                self.boardview.bluearrow = move.cords
                return
        self.boardview.bluearrow = None

    def row_activated(self, widget, *args):
        iter = self.tv.get_selection().get_selected()[1]
        if iter is None:
            return
        board, move, pv = self.store[iter][0]
        # The row may be tied to a specific action.
        path = self.store.get_path(iter)
        indices = path.get_indices()
        if indices:
            self.advisors[indices[0]].row_activated(iter, self.boardview.model)

    def query_tooltip(self, treeview, x, y, keyboard_mode, tooltip):
        # First, find out where the pointer is:
        path_col_x_y = treeview.get_path_at_pos(x, y)

        # If we're not pointed at a row, then return FALSE to say
        # "don't show a tip".
        if not path_col_x_y:
            return False

        # Otherwise, ask the TreeView to set up the tip's area according
        # to the row's rectangle.
        path, col, x, y = path_col_x_y
        if not path:
            return False
        treeview.set_tooltip_row(tooltip, path)

        # And ask the advisor for some text
        indices = path.get_indices()
        if indices:
            text = self.advisors[indices[0]].query_tooltip(path)
            if text:
                label = Gtk.Label()
                label.props.wrap = True
                label.props.width_chars = 60
                label.props.max_width_chars = 60
                label.set_text(text)
                tooltip.set_custom(label)
                # tooltip.set_markup(text)
                return True  # Show the tip.

        return False


################################################################################
# StrengthCellRenderer                                                         #
################################################################################


width, height = 80, 23


class StrengthCellRenderer(Gtk.CellRenderer):
    __gproperties__ = {
        "data": (
            GObject.TYPE_PYOBJECT,
            "Data",
            "Data",
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        ),
    }

    def __init__(self):
        Gtk.CellRenderer.__init__(self)
        self.data = None

    def do_set_property(self, pspec, value):
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
        return getattr(self, pspec.name)

    def do_render(self, context, widget, background_area, cell_area, flags):
        if not self.data:
            return
        text, widthfrac, goodness = self.data
        if widthfrac:
            paintGraph(context, widthfrac, stoplightColor(goodness), cell_area)
        if text:
            layout = PangoCairo.create_layout(context)
            layout.set_text(text, -1)

            fd = Pango.font_description_from_string("Sans 10")
            layout.set_font_description(fd)

            w, h = layout.get_pixel_size()
            context.move_to(cell_area.x, cell_area.y)
            context.rel_move_to(70 - w, (height - h) / 2)

            PangoCairo.show_layout(context, layout)

    def do_get_size(self, widget, cell_area=None):
        return (0, 0, width, height)


GObject.type_register(StrengthCellRenderer)

################################################################################
# StrengthCellRenderer functions                                               #
################################################################################


def stoplightColor(x):
    def interp(y0, yh, y1):
        return y0 + (y1 + 4 * yh - 3 * y0) * x + (-4 * yh + 2 * y0) * x * x

    r = interp(239, 252, 138) / 255
    g = interp(41, 233, 226) / 255
    b = interp(41, 79, 52) / 255
    return r, g, b


def paintGraph(cairo, widthfrac, rgb, rect):
    x, y, w0, h = rect.x, rect.y, rect.width, rect.height
    w = ceil(widthfrac * w0)

    cairo.save()
    cairo.rectangle(x, y, w, h)
    cairo.clip()
    cairo.move_to(x + 10, y)
    cairo.rel_line_to(w - 20, 0)
    cairo.rel_curve_to(10, 0, 10, 0, 10, 10)
    cairo.rel_line_to(0, 3)
    cairo.rel_curve_to(0, 10, 0, 10, -10, 10)
    cairo.rel_line_to(-w + 20, 0)
    cairo.rel_curve_to(-10, 0, -10, 0, -10, -10)
    cairo.rel_line_to(0, -3)
    cairo.rel_curve_to(0, -10, 0, -10, 10, -10)
    cairo.set_source_rgb(*rgb)
    cairo.fill()
    cairo.restore()


# cell renderer for start-stop putton
class CellRendererPixbufXt(Gtk.CellRendererPixbuf):
    __gproperties__ = {
        "active-state": (
            GObject.TYPE_STRING,
            "pixmap/active widget state",
            "stock-icon name representing active widget state",
            None,
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        )
    }
    __gsignals__ = {
        "clicked": (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_STRING,)),
    }

    def __init__(self):
        GObject.GObject.__init__(self)
        self.set_property("mode", Gtk.CellRendererMode.ACTIVATABLE)

    def do_get_property(self, property):
        if property.name == "active-state":
            return self.active_state
        else:
            raise AttributeError("unknown property %s" % property.name)

    def do_set_property(self, property, value):
        if property.name == "active-state":
            self.active_state = value
        else:
            raise AttributeError("unknown property %s" % property.name)

    def do_activate(self, event, widget, path, background_area, cell_area, flags):
        if event.type == Gdk.EventType.BUTTON_PRESS:
            self.emit("clicked", path)

    # def do_clicked(self, path):
    # print "do_clicked", path


GObject.type_register(CellRendererPixbufXt)
